﻿

using System.Numerics;
using System.Security.Cryptography;


namespace exc.jdbi.DiemitkoTheorem;

public partial class DiemitkoPrimeGenerator
{

  public const int PARALLEL_COUNT = 10;
  private readonly static RandomNumberGenerator Rand = RandomNumberGenerator.Create();

  public async static Task<BigInteger[]> StartDiemikoAsync(
    BigInteger halfprime, int size, int pcount = -1)
  {
    var tasks = Enumerable.Range(0, size)
      .Select(x => Task.Run(() => StartDiemitkoSingleAsync(halfprime, pcount))).ToArray();

    var result = await Task.WhenAll(tasks);
    await DisposeAllTaskAsync(tasks);
    return result;
  }

  public async static Task<BigInteger> StartDiemitkoSingleAsync(
    BigInteger halfprime, int pcount = -1)
  {
    var len = halfprime.ToString().Length * 2;
    pcount = pcount < PARALLEL_COUNT || len < 100
               ? PARALLEL_COUNT
               : (int)Math.Sqrt(len);
    return await ToBigPrimeAsync(halfprime, pcount);
  }

  public static (BigInteger Min, BigInteger Max) ToMinMaxFromBits(int bits)
  {
    var _max = (BigInteger.One << bits) - 1;        // 10 bit --> 1111111111 = 1023
    var _min = (BigInteger.One << (bits - 1)) + 1;  // 10 bit --> 1000000000 = 512
    return (_min, _max);
  }

  public async static Task<bool> IsNumericAsync(string[] number) =>
    await Task.Run(() => IsNumeric(number));

  public async static Task<bool> IsNumericAsync(string number) =>
    await Task.Run(() => IsNumeric(number));

  public static bool IsNumeric(string[] number)
  {
    if (number is null || number.Length < 1) return false;
    foreach (var str in number)
      if (!IsNumeric(str))
        return false;
    return true;
  }

  public static bool IsNumeric(string number)
  {
    if (string.IsNullOrEmpty(number)) return false;
    foreach (var digit in number)
      if (!int.TryParse(digit.ToString(), out _))
        return false;
    return true;
  }

  public static bool IsMRPrime(BigInteger bignumber)
  {
    if (bignumber < 2UL) return false;
    if (bignumber == 2UL) return true;
    if (bignumber.IsEven)
      return false; // even number 

    int s = 0;
    BigInteger z = bignumber - 1;
    while ((z & 1) == 0)
    {
      z >>= 1;
      s += 1;
    }

    BigInteger tmp;
    for (int a = 2; a <= 5; a++)
    {
      tmp = BigInteger.ModPow(a, z, bignumber);
      if (tmp == 1 || tmp == bignumber - 1) continue;
      for (int r = 1; r < s; r++)
      {
        tmp = BigInteger.ModPow(tmp, 2, bignumber);
        if (tmp == 1) return false;
        if (tmp == bignumber - 1) break;
      }
      if (!(tmp == bignumber - 1))
        return false;
    }
    return true;
  }

  private async static Task<BigInteger> ToBigPrimeAsync(
    BigInteger halfprime, int pcount = PARALLEL_COUNT) =>
      await ToLevelPrimeAsync(halfprime, pcount);
   

  private async static Task<BigInteger> ToLevelPrimeAsync(
    BigInteger halfprime, int pcount = PARALLEL_COUNT)
  {
    if (pcount < 2) pcount = PARALLEL_COUNT;
    while (true)
    {
      using var cts = new CancellationTokenSource();
      var token = cts.Token;
      var tasks = Enumerable.Range(0, pcount)
          .Select(x => Task.Run(() =>
            DiemitkoPrimeGeneratorAsync(halfprime, token))).ToArray();

      var result = BigInteger.MinusOne;
      var idx = Task.WaitAny(tasks, token);
      if (idx >= 0 && tasks[idx].Result > 1)
      {
        result = tasks[idx].Result;
        cts.Cancel();
      }

      await DisposeAllTaskAsync(tasks);

      if (result > 1) return result;
    }
  } 

  private static BigInteger DiemitkoPrimeGeneratorAsync(
    BigInteger halfprime, CancellationToken token)
  {
    const int a = 2;
    var q = halfprime;
    var r = 4 * (q + 1);

    var qbits = ToNumberOfBits(q);
    var bits = 2 * qbits;

    var pow_a_bits = BigInteger.Pow(a, bits);
    var pow_a_bits_m1 = BigInteger.Pow(a, bits - 1);

    while (true)
    {
      if (token.IsCancellationRequested)
        return -1;

      //step 1: 
      var n1 = pow_a_bits_m1 / q;
      var xi = RngBigInteger(1, pow_a_bits_m1);
      var n2 = pow_a_bits_m1 / xi / q;
      var n = n1 + n2; if ((n & 1) == 1) n++; //n must be even
      if (n >= r) continue;

      //step 2:
      var u = 0ul;

      while (true)
      {
        if (token.IsCancellationRequested)
          return -1;

        //step 3:
        var nu = n + u;
        var p = q * nu + 1;

        //step 4:
        if (p > pow_a_bits) break;

        //step 5:
        if (BigInteger.ModPow(a, BigInteger.Subtract(p, BigInteger.One), p) == 1 &&
            BigInteger.ModPow(a, nu, p) != 1)
          return p;

        u += 2;
      }
    }
  }


  //private static bool CheckTestPrimes(
  //  ref BigInteger candidate, CancellationToken token)
  //{
  //  if ((candidate & 1) == 0) candidate--;
  //  for (int i = 1; i < Test_Primes.Length; i++)
  //  {
  //    if (token.IsCancellationRequested)
  //      return false;

  //    if (candidate == Test_Primes[i])
  //      return true;

  //    if (candidate % Test_Primes[i] == 0)
  //      return false;
  //  }

  //  return true;
  //} 

  private async static Task DisposeAllTaskAsync(
    Task[] tasks)
  {
    try
    {
      await Task.WhenAll(tasks);
      Array.ForEach(tasks, x => x.Dispose());
      tasks = [];
    }
    catch (OperationCanceledException) { }
    catch (Exception) { }
  }

  public static int ToNumberOfBits(BigInteger number) =>
    number.IsPowerOfTwo ? (int)BigInteger.Log2(number) : (int)BigInteger.Log2(number) + 1;

  private static BigInteger RngBigInteger(
    BigInteger min, BigInteger max)
  {
    var bytes = (max - min).ToByteArray();
    Rand.GetNonZeroBytes(bytes);
    bytes[^1] = (byte)(bytes.Last() & 0x7F);

    var result = min + new BigInteger(bytes);
    if (result > max) return RngBigInteger(min, max);
    return result;
  }

  //private static int[] SoELinq(int max)
  //{
  //  int sqrt = (int)Math.Round(Math.Truncate(Math.Sqrt(max)));
  //  var seed = Enumerable.Range(2, max - 1).Select(x => !(x == 2) && (x & 1) == 0 ? 0 : x).ToArray();
  //  var primes = Enumerable.Range(2, sqrt)
  //  .Where(x => (x & 1) == 1)
  //  .Aggregate(seed, (sieve, i) =>
  //  {
  //    if (sieve[i - 2] == 0)
  //      return sieve;
  //    int j = i * i;
  //    while (j <= max)
  //    {
  //      sieve[j - 2] = 0;
  //      j += 2 * i;
  //    }
  //    return sieve;
  //  }).Where(k => !(k == 0));
  //  //File.WriteAllText("p.txt",string.Join(Environment.NewLine, primes));
  //  return primes.ToArray();
  //}


}
